# !/usr/bin/python3
"""
A subgraph object of an OBOGraph object.
"""
from .dag import OboGraph, AbstractNode
import re


class SubGraph(OboGraph):

    """A subgraph of a provided supergraph with node contents."""
    
    def __init__(self, super_graph, namespace_filter=None, allowed_relationships=None):
        """`SubGraph initializer. Creates a subgraph object of :class:`gocats.dag.OboGraph`. Leave `namespace_filter`
        and `allowed_relationship` as :py:obj:`None` to create the entire ontology graph. Otherwise, provide filters to
        limit what information is pulled into the subgraph.

        :param obj super_graph: A supergraph object i.e. :class:`gocats.godag.GoGraph`.
        :param str namespace_filter: Specify the namespace of a sub-ontology namespace, if one is available for the ontology.
        :param list allowed_relationships: Specify a list of relationships to utilize in the graph, other relationships will be ignored.
        """
        self.super_graph = super_graph
        if self.super_graph.namespace_filter and self.super_graph.namespace_filter != namespace_filter:
            raise Exception("Unless a namespace_filter is not specified for a parent_graph, a subgraph's namespace_filter must not differ from its parent graph's namespace_filter.\nsubgraph namespace_filter = {}, supergraph namespace_filter = {}").format(namespace_filter, self.super_graph.namespace_filter)
        if self.super_graph.allowed_relationships and allowed_relationships and any(relationship not in self.super_graph.allowed_relationships for relationship in allowed_relationships):
            raise Exception("Unless an allowed_relationships list is not specified for a parent graph, a subgraph's allowed_relationships list must be a subset of, or exactly, its parent graph's allowed_relationships list.\nsubgraph allowed_relationships = {}, supergraph allowed_relationships = {}").format(allowed_relationships, self.super_graph.allowed_relationships)
        super().__init__(namespace_filter, allowed_relationships)
        self.seeded_size = None  # The number of nodes filtered in the keyword search, used for informational purposes only. 
        self.representative_node = None
        self._root_id_mapping = None
        self._root_node_mapping = None
        self._content_mapping = None

    @property
    def root_id_mapping(self):
        """Property describing a mapping :py:obj:`dict` that relates every ontology term ID of subgraphs in
        :class:`gocats.dag.OboGraph` to a :py:obj:`list` of root, subgraph category-representative node IDs.

        :return: :py:obj:`dict` of :class:`gocats.dag.AbstractNode` IDs mapped to a :py:obj:`list` of root :class:`gocats.dag.AbstractNode` IDs.
        :rtype: :py:obj:`dict`
        """
        if (self._modified and self.representative_node) or self._root_id_mapping is None:
            self._root_id_mapping = {node.id: self.representative_node.id for node in self.representative_node.descendants}
            self._root_id_mapping[self.representative_node.id] = self.representative_node.id
        elif not self.representative_node:
            raise Exception("Mapping failed: top-node not identified.")
        return self._root_id_mapping

    @property
    def root_node_mapping(self):
        """Property describing a mapping :py:obj:`dict` that relates every ontology :class:`gocats.dag.AbstractNode`
        object of subgraphs in :class:`gocats.dag.OboGraph` to a :py:obj:`list` of root, subgraph
        category-representative node objects.

        :return: :py:obj:`dict` of :class:`gocats.dag.AbstractNode` objects mapped to a :py:obj:`list` of root :class:`gocats.dag.AbstractNode` objects.
        :rtype: :py:obj:`dict`
        """
        if (self._modified and self.representative_node) or self._root_node_mapping is None:
            self._root_node_mapping = {node: self.representative_node for node in self.representative_node.descendants}
            self._root_node_mapping[self.representative_node] = self.representative_node
        elif not self.representative_node:
            raise Exception("Mapping failed: top-node not identified.")
        return self._root_node_mapping

    @property
    def content_mapping(self):
        """Property describing a mapping :py:obj:`dict` that relates every root ontology
        :class:`gocats.dag.AbstractNode` IDs of subgraphs in a :class:`gocats.dag.OboGraph` to a :py:obj:`list` of their
        subgraph nodes' IDs.

        :return: :py:obj:`dict` of :class:`gocats.dag.AbstractNode` IDs mapped to a :py:obj:`list' of :class:`gocats.dag.AbstractNode` IDs.
        :rtype: :py:obj:`dict`
        """
        if (self._modified and self.representative_node) or self._content_mapping is None:
            self._content_mapping = {self.representative_node.id: [node.id for node in self.representative_node.descendants]}
            self._content_mapping[self.representative_node.id].append(self.representative_node.id)
        elif not self.representative_node:
            raise Exception("Mapping failed: top-node not identified.")
        return self._content_mapping

    def subnode(self, super_node):
        """Defines a :class:`gocats.subdag.SubGraph` node object. Calls :func:`add_node` to convert a supergraph node
        into a :class:`gocats.subdag.SubGraphNode` and add this node to the subgraph.

        :param super_node: A node object from the supergraph i.e. :class:`gocats.godag.GoGraphNode`.
        :return: A :class:`gocats.subdag.SubGraphNode` object.
        :rtype: :py:obj:`class`
        """
        if super_node.id not in self.id_index:
            self.add_node(super_node)
        return self.id_index[super_node.id]

    def add_node(self, super_node):
        """Converts a supergraph node into a :class:`gocats.subdag.SubGraphNode` and adds this node to the subgraph.
        Sets modification state to :py:obj:`True`.

        :param obj super_node: A node object from the supergraph i.e. :class:`gocats.godag.GoGraphNode`.
        :return: None
        :rtype: :py:obj:`None`
        """
        subgraph_node = SubGraphNode(super_node, self.allowed_relationships)
        if self.valid_node(subgraph_node):
            super().add_node(subgraph_node)
        self._modified = True

    # TODO: Rename/reconsider this (needs to be similar to instantiate_valid_edges)
    def connect_subnodes(self):
        """Analogous to :func:`gocats.dag.instantiate_valid_edges` and :func:`gocats.dag.AbstractEdge.connect_nodes`.
        Updates child and parent node sets for each :class:`gocats.subdag.SubGraphNode` in the
        :class:`gocats.subdag.SubGraph`. Adds edge object references to nodes and node object references to edges.
        Counts instances of relationship IDs and sets modification state to :py:obj:`True`.

        :return: None
        :rtype: :py:obj:`None`
        """
        for subnode in self.node_list:
            subnode.update_children([self.id_index[child.id] for child in subnode.super_node.child_node_set if child.id in self.id_index])
            subnode.update_parents([self.id_index[parent.id] for parent in subnode.super_node.parent_node_set if parent.id in self.id_index])
            for edge in subnode.super_node.edges:  # This counts the number of times each relationship type is used in a subgraph
                if edge.forward_node.id in self.id_index and edge.reverse_node.id in self.id_index:
                    try:
                        self.relationship_count[edge.relationship.id] += 1
                    except KeyError:
                        self.relationship_count[edge.relationship.id] = 1
        self._modified = True

    def greedily_extend_subgraph(self):
        """Extends a seeded subgraph to include all supergraph descendants of the nodes. Searches through the supergraph
        to add new node objects.

        :return: None
        :rtype: :py:obj:`None`
        """
        graph_extension_nodes = set([super_node for super_node in self.super_graph.id_index[self.representative_node.id].descendants if super_node.id not in self.id_index])
        for super_node in graph_extension_nodes:
            self.add_node(super_node)
        self.connect_subnodes()
    
    def conservatively_extend_subgraph(self):
        """Extends a seeded subgraph to include only nodes in the supergraph that occur along paths between nodes in the
        subgraph. Searches through the supergraph to add new node objects.

        :return: None
        :rtype: :py:obj:`None`
        """
        graph_extension_nodes = set()
        for subleaf in self.leaves:
            start_node = self.super_graph.id_index[subleaf.id]
            end_node = self.super_graph.id_index[self.representative_node.id]
            graph_extension_nodes.update(self.nodes_between(start_node, end_node))
        for super_node in graph_extension_nodes:
            if super_node.id not in self.id_index and self.valid_node(super_node):
                self.add_node(super_node)
        self.connect_subnodes()

    def remove_orphan_paths(self):
        """**Not currently in use.**

        Removes nodes and their descendants from the subgraph which do not root to the
        category-representative node.

        :return: None
        :rtype: :py:obj:`None`
        """
        for orphan in self.orphans:
            orphaned_descendants = orphan.descendants - self.representative_node.descendants
            if orphaned_descendants:
                for descendant in orphaned_descendants:
                    self.remove_node(descendant)
            self.remove_node(orphan)

    @staticmethod
    def find_representative_node(subgraph, search_string_list):
        """Extracts candidate :class:`gocats.subdag.SubGraphNode` objects from the :class:`gocats.subdag.SubGraph`
        objects based on a list of search strings matching strings in the names of the nodes (using regular
        expressions). Returns the candidate node with the highest number of descendants. Returns the sole node if the
        subgraph only contains one node, aborts if the subgraph is empty.

        :param subgraph: A :class:`gocats.subdag.SubGraph` object.
        :param search_string_list: A :py:obj:`list` of search term :py:obj:`str` entries.
        :return: A candidate term :class:`gocats.subgraph.SubGraphNode` chosen as the subgraph's representative ontology term.
        """
        if len(subgraph.node_list) == 1:
            return subgraph.node_list[0]
        elif not subgraph.node_list:
            raise Exception("Subgraph did not seed any nodes from the supergraph! Aborting.")
        else:
            candidates = [node for node in subgraph.node_list if any(re.search('(?<!\-)'+search_string+'(?!\-)', node.name) for search_string in search_string_list) and node not in subgraph.leaves and not node.obsolete] 
            representative_node_scoring = {node: len(node.descendants) for node in candidates}
            return max(representative_node_scoring, key=representative_node_scoring.get)

    @staticmethod
    def from_filtered_graph(super_graph, keyword_list, namespace_filter=None, allowed_relationships=None, extension='greedy'):
        """Staticmethod for extracting a subgraph from the supergraph by selecting nodes that contain vocabulary in the
        supplied keyword list. Leave `namespace_filter` and `allowed_relationship` as :py:obj:`None` to create the
        entire ontology graph. Otherwise, provide filters to limit what information is pulled into the subgraph.
        Graph `extension` variable defaults to 'greedy' which calls :func:`greedily_extend_subgraph` to add nodes to the
        subgraph after instantiation. Conversely, 'conservative' may be used to call
        :func:`conservatively_extend_subgraph` for this function.

        :param obj super_graph: A supergraph object i.e. :class:`gocats.godag.GoGraph`.
        :param keyword_list: A :py:obj:`list` of :py:obj:`str` entries used to query the supergraph for concepts to be extracted into subgraphs.
        :param str namespace_filter: Specify the namespace of a sub-ontology namespace, if one is available for the ontology.
        :param list allowed_relationships: Specify a list of relationships to utilize in the graph, other relationships will be ignored.
        :param str extension: Specify 'greedy' or 'conservative' to determine how subgraphs will be extended after creation (defaults to greedy).
        :return: A :class:`gocats.subdag.SubGraph` object.
        """
        subgraph = SubGraph(super_graph, namespace_filter, allowed_relationships)
        keyword_list = [word.lower() for word in keyword_list]
        filtered_nodes = super_graph.filter_nodes(keyword_list)
        subgraph.seeded_size = len(filtered_nodes)
        for super_node in filtered_nodes:
            subgraph.add_node(super_node)
        subgraph.connect_subnodes()
        subgraph.representative_node = subgraph.find_representative_node(subgraph, keyword_list)
        subgraph.root_nodes.append(subgraph.representative_node)
        if extension == 'greedy':
            subgraph.greedily_extend_subgraph()
        elif extension == 'conservative':
            subgraph.conservatively_extend_subgraph()
        subgraph_orphans_descendants = set()
        for orphan in subgraph.orphans:
            for node in orphan.descendants:
                subgraph_orphans_descendants.add(node)
        subgraph_orphans_descendants.update([orphan for orphan in subgraph.orphans])
            
        return subgraph


class SubGraphNode(AbstractNode):

    """An instance of a node within a subgraph of an OBO ontology (supergraph)
    """
    
    def __init__(self, super_node, allowed_relationships=None):
        """SubGraphNode initializer. Inherits from :class:`gocats.dag.AbstractNode` and contains a reference to the
        supergraph node it represents e.g. :class:`gocats.godag.GoGraphNode`.

        :param super_node: A node from the `supergraph`.
        :param allowed_relationships: **Not currently used** Used to specify a list of allowable relationships evaluated between nodes.
        """
        self.super_node = super_node
        self.parent_node_set = set()
        self.child_node_set = set()
        self._modified = True
        self._descendants = None
        self._ancestors = None

    # TODO: add in update_parent_node_set and update_child_node_set with a _modified switch !!!!

    @property
    def super_edges(self):
        """:py:obj:`property` describing the set of edges referenced in the supergraph node, filtered to only those
         edges with nodes in the subgraph node.

        :return: A set of :class:`gocats.subgraph.SubGraphNode` edges that were copied from the supergraph node.
        :rtype: :py:obj:`set`
        """
        edges = set()
        edges.add([edge for edge in self.super_node.edges if edge.parent_node.id in [node.id for node in self.parent_node_set] and edge.child_node.id in [node.id for node in self.child_node_set]])
        return edges

    @property
    def id(self):
        """:py:obj:`property` describing the ID of the supernode

        :return: The ID of a supernode e.g. :class:`gocats.godag.GoGraphNode`
        :rtype: :py:obj:`str`
        """
        return self.super_node.id

    @property
    def name(self):
        """:py:obj:`property` describing the name of the supernode

        :return: The name of a supernode e.g. :class:`gocats.godag.GoGraphNode`
        :rtype: :py:obj:`str`
        """
        return self.super_node.name
    
    @property
    def definition(self):
        """:py:obj:`property` describing the definition of the supernode

        :return: The definition of a supernode e.g. :class:`gocats.godag.GoGraphNode`
        :rtype: :py:obj:`str`
        """
        return self.super_node.definition

    @property
    def namespace(self):
        """:py:obj:`property` describing the namespace of the supernode

        :return: A namespace of a supernode e.g. :class:`gocats.godag.GoGraphNode`
        :rtype: :py:obj:`str`
        """
        return self.super_node.namespace

    @property
    def obsolete(self):
        """:py:obj:`property` describing whether or not supernode is marked as obsolete.

        :return: :py:obj:`True` or :py:obj:`False`
        """
        return self.super_node.obsolete

    def update_parents(self, parent_set):
        """Updates the parent_node_set with a set of new parents provided. Sets modification state to :py:obj:`True`.

        :param parent_set: A set of parent nodes to be added to this objects parent_node set.
        :return: None
        :rtype: :py:obj:`None`
        """
        self.parent_node_set.update(parent_set)
        self._modified = True

    def update_children(self, child_set):
        """Updates the child_node_set with a set of new children provided. Sets modification state to :py:obj:`True`.

        :param child_set: A set of child nodes to be added to this objects child_node set.
        :return: None
        :rtype: :py:obj:`None`
        """
        self.child_node_set.update(child_set)
        self._modified = True



#!/usr/bin/env python3

"""
The Gene Ontology Categories Suite (GOcats)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides methods for the creation of directed acyclic concept subgraphs of Gene Ontology, along with methods
for evaluating those subgraphs.

Command line implementation::

    Usage:
        gocats create_subgraphs <database_file> <keyword_file> <output_directory> [--supergraph_namespace=<None> --subgraph_namespace=<None> --supergraph_relationships=[] --subgraph_relationships=[] --map_supersets --output_termlist]
        gocats categorize_dataset <dataset_file> <term_mapping> <output_directory> <mapped_dataset_filename> [--dataset_type=<GAF> --entity_col=<0> --go_col=<1> --retain_unmapped_annotations]

    Options:
        -h | --help                          Shows this screen.
        --version                            Shows version.
        --supergraph_namespace=<None>        Filters the supergraph to a given namespace.
        --subgraph_namespace=<None>          Filters the subgraph to a given namespace.
        --supergraph_relationships=[]        A provided list will denote which relationships are allowed in the supergraph.
        --subgraph_relationships=[]          A provided list will denote which relationships are allowed in the subgraph.
        --map_supersets                      Maps all terms to all root nodes, regardless of if a root node subsumes another.
        --output_termlist                    Outputs a list of all terms in the supergraph as a JsonPickle file in the output directory.
        --dataset_type=<GAF>                 Enter file type for dataset [GAF|TSV|CSV]. Defaults to GAF.
        --entity_col=<0>                     If CSV or TSV file type, indicate which column the entity IDs are listed. Defaults to 0.
        --go_col=<1>                         If CSV or TSV file type, indicate which column the GO IDs are listed. Defaults to 1.
        --retain_unmapped_annotations        If specified, annotations that are not mapped to a concept are copied into the mapped dataset output file with its original annotation.
"""

import docopt
from . import gocats
from ._version import __version__


def main(args):
    if args['create_subgraphs']:
        gocats.create_subgraphs(args)
    elif args['categorize_dataset']:
        gocats.categorize_dataset(args)

if __name__ == '__main__':
    args = docopt.docopt(__doc__, version=str('GOcats Version ') + __version__)
    main(args)



# !/usr/bin/python3
"""
The Gene Ontology Categories Suite (GOcats)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides methods for the creation of directed acyclic concept subgraphs of Gene Ontology, along with methods
for evaluating those subgraphs.
"""
import os
import sys
import re
import csv
from . import ontologyparser
from . import godag
from . import subdag
from . import tools


# Need a SubGraphCollection object
def build_graph(args):
    """**Not yet implemented**

    Try build_graph_interpreter to create a GO graph object to explore within a Python interpreter."""
    # FIXME: JsonPickle is reaching max recursion depth because of the fact that objects point to one another.
    if args['--supergraph_namespace']:
        supergraph_namespace = args['--supergraph_namespace']
    else:
        supergraph_namespace = None
    if args['--allowed_relationships']:
        allowed_relationships = args['--allowed_relationships']
    else:
        allowed_relationships = None

    database = open(args['<database_file>'], 'r')
    output_directory = args['<output_directory>']
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    database_name = os.path.basename(args['<database_file>'])
    graph_class = {'go.obo': godag.GoGraph(supergraph_namespace, allowed_relationships)}
    graph = graph_class[database_name]
    parsing_class = {'go.obo': ontologyparser.GoParser(database, graph)}
    parsing_class[database_name].parse()
    database.close()


def build_graph_interpreter(database_file, supergraph_namespace=None, allowed_relationships=None, relationship_directionality='gocats'):
    """Creates a graph object of GO, which can be traversed and queried within a Python interpreter.

    :param file_handle database_file: Ontology database file.
    :param str supergraph_namespace: Optional - Filter graph to a sub-ontology namespace.
    :param list allowed_relationships: Optional - Filter graph to use only those relationships listed.
    :param relationship_directionality: Optional - Any string other than 'gocats' will retain all original GO relationship directionalities. Defaults to reverseing has_part direction.
    :return: A Graph object of the ontology provided.
    :rtype: :py:obj:`class`
    """
    database = open(database_file, 'r')
    graph = godag.GoGraph(supergraph_namespace, allowed_relationships)
    go_parser = ontologyparser.GoParser(database, graph, relationship_directionality=relationship_directionality)
    go_parser.parse()
    database.close()
    return graph


def create_subgraphs(args):
    """Creates a graph object of an ontology, processed into :class:`gocats.dag.OboGraph` or to an object that
    inherits from :class:`gocats.dag.OboGraph`, and then extracts subgraphs which represent concepts that are defined
    by a list of provided keywords. Each subgraph is processed into :class:`gocats.subdag.SubGraph`.

    :param database_file: Ontology database file.
    :param keyword_file: A CSV file with two columns: column 1 naming categories, and column 2 listing search strings (no quotation marks, separated by semicolons).
    :param output_directory: The directory where results are stored.
    :param --supergraph_namespace=<None>: OPTIONAL-Specify a supergraph sub-ontology to filter e.g. cellular_component.
    :param --subgraph_namespace=<None>: OPTIONAL-Specify a subgraph sub-ontology to filter e.g. cellular_component.
    :param --supergraph_relationships=[]: OPTIONAL-Specify a list of relationships to limit in the supergraph e.g. [is_a, part_of].
    :param --subgraph_relationships=[]: OPTIONAL-Specify a list of relationships to limit in subgraphs e.g. [is_a, part_of].
    :param --map_supersets: OPTIONAL-Allow subgraphs to subsume other subgraphs.
    :param --output_termlist: OPTIONAL-Create a translation of ontology terms to their names to improve interpretability of dev test results.
    :param --go-basic-scoping: OPTIONAL-Creates a GO graph similar to go-basic with only scoping-type relationships (is_a and part_of).
    :param --network_table_name=<None>: OPTIONAL-Make a specific name for the network table produced from the subgraphs (defaults to NetworkTable.csv)
    :return: None
    :rtype: :py:obj:`None`
    """

    if args['--supergraph_namespace']:
        supergraph_namespace = args['--supergraph_namespace']
    else:
        supergraph_namespace = None
    if args['--subgraph_namespace']:
        subgraph_namespace = args['--subgraph_namespace']
    else:
        subgraph_namespace = None
    if args['--supergraph_relationships']:
        supergraph_relationships = args['--supergraph_relationships']
    else:
        supergraph_relationships = None
    if args['--subgraph_relationships']:
        subgraph_relationships = args['--subgraph_relationships']
    else:
        subgraph_relationships = None
    if args['--go-basic-scoping']:
        supergraph_relationships = ['is_a', 'part_of']
        subgraph_relationships = ['is_a', 'part_of']

    # Building the supergraph
    database = open(args['<database_file>'], 'r')
    output_directory = args['<output_directory>']
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    database_name = os.path.basename(args['<database_file>'])
    graph_class = {'go.obo': godag.GoGraph(supergraph_namespace, supergraph_relationships)}
    try:
        supergraph = graph_class[database_name]
    except KeyError:
        print("The provided ontology filename was not recognized. Please do not rename ontology files. The accepted list of file names are as follows: \n", graph_class.keys())
        sys.exit()
    parsing_class = {'go.obo': ontologyparser.GoParser(database, supergraph)}
    try:
        parsing_class[database_name].parse()
    except KeyError:
        print("The provided ontology filename was not recognized. Please do not rename ontology files. The accepted list of file names are as follows: \n", graph_class.keys())
        sys.exit()
    if args['--output_termlist']:
        tools.jsonpickle_save(list(supergraph.id_index.keys()), os.path.join(args['<output_directory>'], "termlist"))

    id_translation = dict()
    for id, node in supergraph.id_index.items():
        id_translation[id] = node.name
    tools.jsonpickle_save(id_translation, os.path.join(args['<output_directory>'], "id_translation"))

    database.close()

    # Building and collecting subgraphs
    subgraph_collection = dict()
    with open(args['<keyword_file>'], newline='') as file:
        reader = csv.reader(file, delimiter=',', quoting=csv.QUOTE_MINIMAL)
        for row in reader:
            subgraph_name = row[0]
            keyword_list = [keyword for keyword in re.split(';', row[1])]
            subgraph_collection[subgraph_name] = subdag.SubGraph.from_filtered_graph(supergraph, keyword_list, subgraph_namespace, subgraph_relationships)

    # Handling superset mapping
    if not args['--map_supersets']:
        category_subsets = find_category_subsets(subgraph_collection)
    else:
        print("NOTE: supersets were mapped.")
        category_subsets = None

    collection_id_mapping = dict()
    collection_node_mapping = dict()
    collection_content_mapping = dict()
    for subgraph in subgraph_collection.values():
        for node_id, rep_node in subgraph.root_id_mapping.items():
            try:
                collection_id_mapping[node_id].update([rep_node])
            except KeyError:
                collection_id_mapping[node_id] = set([rep_node])
        for node, rep_node in subgraph.root_node_mapping.items():
            try:
                collection_node_mapping[node].update(rep_node)
            except KeyError:
                collection_node_mapping[node] = set([rep_node])
        for rep_node, content in subgraph.content_mapping.items():
            collection_content_mapping[rep_node] = content

    # Remove root nodes that are subsets of existing root nodes from mapping
    if category_subsets:
        for node_id, root_id_list in collection_id_mapping.items():
            for subset_id, superset_ids in category_subsets.items():
                if subset_id in root_id_list:
                    [root_id_list.remove(node) for node in superset_ids if node in root_id_list]
    # TODO: do the same for node_object_mapping

    # Save mapping files and create report
    tools.jsonpickle_save(collection_id_mapping, os.path.join(args['<output_directory>'], "GC_id_mapping"))
    tools.json_save(collection_id_mapping, os.path.join(args['<output_directory>'], "GC_id_mapping"))
    tools.jsonpickle_save(collection_content_mapping, os.path.join(args['<output_directory>'], "GC_content_mapping"))
    tools.json_save(collection_content_mapping, os.path.join(args['<output_directory>'], "GC_content_mapping"))
    with open(os.path.join(output_directory, 'subgraph_report.txt'), 'w') as report_file:
        report_file.write(
            'Subgraph data\nSupergraph filter: {}\nSubgraph filter: {}\nGO terms in the supergraph: {}\nGO terms in subgraphs: {}\nRelationship prevalence: {}'.format(
                supergraph_namespace, subgraph_namespace, len(set(supergraph.node_list)),
                len(set(collection_id_mapping.keys())), supergraph.relationship_count))
        for subgraph_name, subgraph in subgraph_collection.items():
            out_string = """
                -------------------------
                {}
                Subgraph relationships: {}
                Seeded size: {}
                Representative node: {}
                Nodes added: {}
                Non-subgraph hits (orphans): {}
                Total nodes: {}
                """.format(subgraph_name, subgraph.relationship_count, subgraph.seeded_size,
                           subgraph.representative_node.name, len(subgraph.node_list) - subgraph.seeded_size,
                           len(subgraph.node_list) - len(subgraph.root_id_mapping.keys()),
                           len(subgraph.root_node_mapping.keys()))
            report_file.write(out_string)

    # FIXME: cannot json save due to recursion of objects within objects...
    # tools.jsonpickle_save(collection_node_mapping, os.path.join(args['<output_directory>'], "GC_node_mapping"))

    if args['network_table_name']:
        network_table_name = args['network_table_name']
    else:
        network_table_name = "Network_table.csv"

    # Making a file for network visualization via Cytoscape 3.0
    with open(os.path.join(args['<output_directory>'], network_table_name), 'w', newline='') as network_table:
        edgewriter = csv.writer(network_table, delimiter=',', quotechar='|', quoting=csv.QUOTE_MINIMAL)
        for term_id, root_id_list in collection_id_mapping.items():
            for root_id in root_id_list:
                edgewriter.writerow([term_id, root_id])


def find_category_subsets(subgraph_collection):
    """Finds subgraphs which are subsets of other subgraphs to remove redundancy, when specified.

    :param subgraph_collection: A dictionary of subgraph objects (keys: subgraph name, values: subgraph object).
    :return: A dictionary relating which subgraph objects are subsets of other subgraphs (keys: subset subgraph, values: superset subgraphs).
    :rtype: :py:obj:`dict`
    """
    is_subset_of = dict()
    for subgraph in subgraph_collection.values():
        for next_subgraph in subgraph_collection.values():
            if subgraph.representative_node.id != next_subgraph.representative_node.id and subgraph.representative_node.id in next_subgraph.root_id_mapping.keys():
                try:
                    is_subset_of[subgraph.representative_node.id].add(next_subgraph.representative_node.id)
                except KeyError:
                    is_subset_of[subgraph.representative_node.id] = {next_subgraph.representative_node.id}
    return is_subset_of


def categorize_dataset(args):
    """Reads in a Gene Annotation File (GAF) and maps the annotations contained therein to the categories organized by
    GOcats or other methods. Outputs a mapped GAF and a list of unmapped genes in the specified output directory.

    :param dataset_file: A file containing gene annotations.
    :param term_mapping: A dictionary mapping category-defining ontology terms to their subgraph children terms. May be produced by GOcats or another method.
    :param output_directory: Specify the directory where the output file will be stored.
    :param mapped_dataset_filename: Specify the desired name of the mapped GAF.
    :param --dataset_type: Enter file type for dataset [GAF|TSV|CSV]. Defaults to GAF.
    :param --entity_col=<0>: If CSV or TSV file type, indicate which column the entity IDs are listed. Defaults to 0.
    :param --go_col: If CSV or TSV file type, indicate which column the GO IDs are listed. Defaults to 1.
    :param --retain_unmapped_annotations: If specified, annotations that are not mapped to a concept are copied into the mapped dataset output file with its original annotation.
    :return: None
    :rtype: :py:obj:`None`
    """
    if args['--dataset_type']:
        dataset_type = args['--dataset_type']
    else:
        dataset_type = None
    if args['--entity_col']:
        entity_id_index = int(args['--entity_col'])
    else:
        entity_id_index = 0
    if args['--go_col']:
        go_id_index = int(args['--go_col'])
    else:
        go_id_index = 1

    mapping_dict = tools.jsonpickle_load(args['<term_mapping>'])
    output_directory = os.path.realpath(args['<output_directory>'])
    mapped_dataset_filename = args['<mapped_dataset_filename>']
    unmapped_entities = set()
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    if dataset_type == "GAF" or not dataset_type:
        loaded_gaf_array = tools.parse_gaf(args['<dataset_file>'])
        mapped_gaf_array = list()
        for line in loaded_gaf_array:
            if line[4] in mapping_dict.keys():
                mapped_terms = mapping_dict[line[4]]
                for term in mapped_terms:
                    mapped_gaf_array.append(line[0:4] + [term] + line[5:-1])
            else:
                if line[2] == '':
                    unmapped_entities.add('NO_GENE:' + line[1])
                else:
                    unmapped_entities.add(line[2])
        tools.write_out_gaf(mapped_gaf_array, os.path.join(output_directory, mapped_dataset_filename))
        tools.list_to_file(os.path.join(output_directory, mapped_dataset_filename + '_unmappedEntities'), unmapped_entities)

    elif dataset_type == "CSV":
        mapped_rows = []
        with open(os.path.realpath(args['<dataset_file>']), 'r') as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=',')
            header = next(csv_reader)
            for row in csv_reader:
                mapped_row = row
                if mapped_row[go_id_index] in mapping_dict.keys():
                    for concept_term in mapping_dict[mapped_row[go_id_index]]:
                        mapped_row[go_id_index] = concept_term
                        mapped_rows.append(mapped_row)
                else:
                    unmapped_entities.add(mapped_row[entity_id_index])
                    if args['--retain_unmapped_annotations']:
                        mapped_rows.append(mapped_row)
            mapped_rows.insert(0, header)
        with open(os.path.join(output_directory, mapped_dataset_filename), 'w') as output_csv:
            csv_writer = csv.writer(output_csv, delimiter=',')
            for row in mapped_rows:
                csv_writer.writerow([item for item in row])
        tools.list_to_file(os.path.join(output_directory, 'unmappedEntities'), unmapped_entities)
